from functools import partial
from uuid import UUID

import anyio
from loguru import logger
from typing_extensions import Self

from .._internal.task_scheduler import TaskScheduler
from .._version import VERSION
from ..config import CONFIG, AgentConfig
from . import host, worker
from .client import AgentClient
from .schemas import Heartbeat, LoginPayload
from .service_loop import critical_service_loop


class Agent:
    _config: AgentConfig
    client: AgentClient

    def __init__(self, config: AgentConfig):
        self.set_config(config)

    @classmethod
    def default(cls) -> Self:
        return cls(CONFIG)

    @classmethod
    def from_config_file(cls, filename: str) -> Self:
        config = AgentConfig.load(filename)
        return cls(config)

    @property
    def config(self) -> AgentConfig:
        return self._config

    def set_config(self, value: AgentConfig):
        self._config = value
        self.client = AgentClient(value)

    @property
    def id(self) -> UUID:
        return self.config.agent_id

    @property
    def has_logged_in(self) -> bool:
        return self.config.is_valid() and self.config.logged_in

    async def login(self, encoded_token: str):
        self.config.set_auth_token(encoded_token)
        # update the client with the new config (server_url and token)
        self.client.set_config(self.config)
        await self.report_host_info()
        self.config.logged_in = True
        self.config.save()

    async def report_host_info(self):
        hostname, ip_address = host.get_hostname_ip()
        payload = LoginPayload(
            agent_id=self.id,
            hostname=hostname,
            ip_address=ip_address,
            agent_version=VERSION,
            host_info=host.get_host_info(),
        )
        await self.client.login(payload)

    async def logout(self):
        await self.client.logout(agent_id=self.id)
        self.config.clear_auth_token()
        self.config.logged_in = False
        self.client.set_config(self.config)
        self.config.save()

    async def send_heartbeat(self):
        path = worker.get_docker_root_dir()
        payload = Heartbeat(
            agent_id=self.id,
            metrics=host.get_host_metrics(path=path),
            service_status=worker.get_service_status(),
        )
        await self.client.heartbeat(payload)

    async def sync_with_server(self):
        try:
            await self.send_heartbeat()
        except Exception as e:
            logger.error(f"Failed to sync with server: {e}")

    async def start(self):
        scheduler = TaskScheduler(self)
        try:
            await self.report_host_info()
        except Exception as e:
            logger.debug(f"Failed to report host info: {e}")

        async with anyio.create_task_group() as tg:
            # wait for an initial heartbeat to configure the worker
            await self.sync_with_server()

            logger.info("Start sending heartbeats.")
            tg.start_soon(
                partial(
                    critical_service_loop,
                    workload=self.sync_with_server,
                    interval=self.config.heartbeat_interval,
                    printer=logger.info,
                    jitter_range=0.3,
                    backoff=4,
                )
            )

            logger.info("Start polling for new tasks.")
            tg.start_soon(
                partial(
                    critical_service_loop,
                    workload=scheduler.get_and_dispatch_tasks,
                    interval=self.config.poll_interval,
                    printer=logger.info,
                    backoff=4,
                )
            )



import asyncio
import base64
import json
import subprocess
import sys
import tempfile
import time
from pathlib import Path
from typing import TYPE_CHECKING

from loguru import logger

from .schemas import TaskItem, TaskStatus, UpdateTaskInstanceStatus
from .utils import utcnow

if TYPE_CHECKING:
    from .client import AgentClient


class TaskExecutor:
    def __init__(
        self,
        client: "AgentClient",
        task_heartbeat_interval: int = 5,
        task_retry_interval: int = 5,
    ):
        self.client = client
        self.task_heartbeat_interval = task_heartbeat_interval
        self.task_retry_interval = task_retry_interval

    async def execute(self, task: "TaskItem") -> None:
        start_time = time.time()
        with tempfile.TemporaryDirectory() as temp_dir:
            await self.client.update_task_status(
                task.task_instance_id,
                UpdateTaskInstanceStatus(
                    status=TaskStatus.RUNNING,
                    logs=[f"{utcnow()} | Task {task.task_instance_id} started in working directory: {temp_dir}"],
                ),
            )

            logger.info(f"Task {task.task_instance_id} started in working directory: {temp_dir}")
            await self._execute_in_tempdir(task, temp_dir)

        logger.info(f"Task {task.task_instance_id} completed in {time.time() - start_time:.2f} seconds.")

    async def _execute_in_tempdir(self, task: "TaskItem", workdir: str):
        task_instance_id = task.task_instance_id
        retries = 0

        cmd_args, output_file = self._prepare_execution_context(task, workdir)

        while retries <= task.max_retries:
            task_event = asyncio.Event()
            task_event.clear()
            log_queue = asyncio.Queue()

            # the heartbeat task keeps sending the task status and logs to the server
            heartbeat_task = asyncio.create_task(self.send_heartbeat(task_instance_id, task_event, log_queue))

            try:
                logger.info(f"Executing task: {task_instance_id} (Attempt {retries + 1})")
                await log_queue.put(f"{utcnow()} | Attempt {retries + 1}")

                await self.execute_in_subprocess(
                    *cmd_args,
                    cwd=workdir,
                    log_queue=log_queue,
                    timeout=task.max_duration,
                )

                # success, read result and update status
                result = None
                if output_file.exists():
                    result = json.loads(output_file.read_text())

                logger.info(f"Task {task_instance_id} completed successfully.")

                await self.client.update_task_status(
                    task_instance_id,
                    UpdateTaskInstanceStatus(
                        status=TaskStatus.SUCCESS,
                        result=result,
                        logs=await self.drain_queue(log_queue),
                    ),
                )
                break
            except (subprocess.CalledProcessError, TimeoutError) as e:
                retries += 1
                logger.error(f"Attempt {retries} failed: {e}")
                await log_queue.put(f"{utcnow()} | Attempt {retries} failed: {e}")
                if retries > task.max_retries:
                    await self.client.update_task_status(
                        task_instance_id,
                        UpdateTaskInstanceStatus(
                            status=TaskStatus.FAILED,
                            logs=await self.drain_queue(log_queue),
                        ),
                    )
                    task_event.set()
                    await heartbeat_task
                    return

                await asyncio.sleep(self.task_retry_interval)

            finally:
                task_event.set()
                await heartbeat_task

    def _prepare_execution_context(self, task: "TaskItem", workdir: str) -> tuple[tuple[str, ...], Path]:
        script = Path(workdir) / f"handler.{task.handler_format}"
        script.write_bytes(base64.b64decode(task.handler_code))

        input_file = Path(workdir) / "payload.json"
        input_file.write_text(json.dumps(task.payload, indent=2))

        output_file = Path(workdir) / "result.json"

        # python /path/to/handler.py --input /path/to/payload.json --output /path/to/result.json
        cmd_args = (
            str(script),
            "--input",
            str(input_file),
            "--output",
            str(output_file),
        )
        return cmd_args, output_file

    async def execute_in_subprocess(self, *args: str, cwd: str, log_queue: asyncio.Queue[str], timeout: int):
        process = await asyncio.create_subprocess_exec(
            sys.executable,
            *args,
            cwd=cwd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        try:
            await asyncio.wait_for(
                asyncio.gather(
                    self.read_stream(process.stdout, log_queue),
                    self.read_stream(process.stderr, log_queue),
                ),
                timeout=timeout,
            )
        except asyncio.TimeoutError:
            logger.error(f"Subprocess timed out after {timeout} seconds, killing it.")
            process.kill()
            await process.wait()
            raise TimeoutError(f"Subprocess timed out after {timeout} seconds")

        rc = await process.wait()
        if rc != 0:
            raise subprocess.CalledProcessError(rc, "python")

    async def read_stream(self, stream: asyncio.StreamReader, log_queue: asyncio.Queue[str]):
        while True:
            line = await stream.readline()
            if not line:
                break
            await log_queue.put(line.decode().strip())

    async def send_heartbeat(self, task_instance_id: int, task_event: asyncio.Event, log_queue: asyncio.Queue):
        logger.info(f"Task {task_instance_id} heartbeat started.")
        while not task_event.is_set():
            logs: list[str] = []
            while not log_queue.empty():
                logs.append(await log_queue.get())

            payload = UpdateTaskInstanceStatus(
                status=TaskStatus.RUNNING,
                logs=logs,
            )
            await self.client.update_task_status(task_instance_id, payload)
            await asyncio.sleep(self.task_heartbeat_interval)

        logger.info(f"Task {task_instance_id} heartbeat stopped.")

    async def drain_queue(self, log_queue: asyncio.Queue[str]) -> list[str]:
        logs: list[str] = []
        while not log_queue.empty():
            logs.append(await log_queue.get())
        return logs



from __future__ import annotations

import datetime
from enum import Enum
from typing import Any, Generic, Optional, TypeVar
from uuid import UUID

from pydantic import BaseModel, Field

from .host import HostInfo, HostMetrics
from .utils import utcnow

T = TypeVar("T")


class RecurveEnum(str, Enum):
    """Base Enum class for Recurve."""

    def __str__(self) -> str:
        return str.__str__(self)


class ServiceStatusInfo(BaseModel):
    container_name: str
    status: str


class LoginPayload(BaseModel):
    agent_id: UUID
    ip_address: str
    hostname: str
    agent_version: str
    host_info: HostInfo


class Heartbeat(BaseModel):
    agent_id: UUID
    metrics: HostMetrics
    sent_time: datetime.datetime = Field(default_factory=utcnow)
    service_status: list[ServiceStatusInfo] = Field(default_factory=list)


class Pagination(BaseModel, Generic[T]):
    total: int
    items: list[T]


class TaskItem(BaseModel):
    task_instance_id: int
    agent_id: UUID
    queue: str
    max_retries: int
    max_duration: int
    handler_code: str  # base64 encoded content of the handler file
    handler_format: str  # format of the handler file, either "py" or "zip"
    payload: dict  # payload for the task, will be passed to the handler in JSON format


class TaskStatus(RecurveEnum):
    PENDING = "pending"
    RUNNING = "running"
    SUCCESS = "success"
    FAILED = "failed"


class UpdateTaskInstanceStatus(BaseModel):
    status: TaskStatus
    logs: Optional[list[str]] = None
    result: Optional[dict[str, Any]] = None



import asyncio
from typing import TYPE_CHECKING

from loguru import logger

if TYPE_CHECKING:
    from .agent import Agent

from .task_executor import TaskExecutor


class TaskScheduler:
    def __init__(self, agent: "Agent"):
        self.agent = agent
        self.task_executor = TaskExecutor(agent.client)

        self._last_schedule_failed = False

    async def get_and_dispatch_tasks(self):
        try:
            paged_tasks = await self.agent.client.get_pending_tasks(agent_id=self.agent.id)
            if self._last_schedule_failed:
                logger.info("Connection to server is restored.")
                self._last_schedule_failed = False

            if paged_tasks.total > 0:
                logger.info(f"Got {paged_tasks.total} pending tasks, sending them to executor.")
            for task in paged_tasks.items:
                # fire and forget, let the task executor handle the task
                asyncio.create_task(self.task_executor.execute(task))
        except Exception as e:
            logger.error(f"Failed to get pending tasks: {e}")
            self._last_schedule_failed = True



import platform
import sys
from enum import Enum

import typer
from loguru import logger

from ._typer import RecurveTyper, exit_with_error
from .agent import agent_app
from .config import config_app

app = RecurveTyper(help="Recurve Agent Command Line Interface")
app.add_typer(agent_app, name="agent")
app.add_typer(config_app, name="config")


class LogLevel(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


@app.callback()
def on_init(log_level: LogLevel = LogLevel.INFO):
    logger.remove()

    logger.add(sys.stderr, level=log_level)


@app.command()
async def version():
    """Show version information."""
    from recurvedata.agent._version import VERSION

    py_impl = platform.python_implementation()
    py_version = platform.python_version()
    system = platform.system()

    typer.echo(f"Running Recurve Agent {VERSION} with {py_impl} {py_version} on {system}.")


@app.command()
async def login():
    """Login the agent to the Recurve service."""
    from .._internal.agent import Agent

    agent = Agent.default()
    if agent.has_logged_in:
        yes: bool = typer.confirm("Agent already logged in. Do you want to re-login?", default=False)
        if not yes:
            return

    encoded_token: str = typer.prompt("Paste your API key", hide_input=True)

    try:
        await agent.login(encoded_token)
    except ValueError:
        exit_with_error("Invalid token.")
    except Exception as e:
        exit_with_error(f"Failed to login: {e}")

    typer.secho("Agent logged in.", fg="green")


@app.command()
async def logout():
    """Logout the agent from the Recurve service."""
    from .._internal.agent import Agent

    agent = Agent.default()
    if not agent.has_logged_in:
        exit_with_error("Agent not logged in.", fg=typer.colors.YELLOW)

    await agent.logout()


if __name__ == "__main__":
    app()



from __future__ import annotations

import base64
import json
import os
import uuid
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, get_origin

from pydantic import BaseModel, ConfigDict, Field, SecretStr
from typing_extensions import Self

if TYPE_CHECKING:
    from pydantic.fields import FieldInfo

RECURVE_HOME = Path(os.environ.get("RECURVE_HOME", Path.home() / ".recurve"))
CONFIG_FILE_PATH = RECURVE_HOME / "config.json"


class AgentConfig(BaseModel):
    model_config = ConfigDict(extra="ignore")
    editable_fields: ClassVar[set[str]] = {"token", "heartbeat_interval", "poll_interval", "request_timeout"}

    agent_id: uuid.UUID = Field(default_factory=uuid.uuid4, description="The unique identifier of the agent.")
    server_host: str = Field(..., description="The hostname of the server.")
    tenant_domain: str = Field(..., description="The domain of the tenant.")
    token: SecretStr = Field(..., description="The authentication token.")
    logged_in: bool = Field(False, description="Whether the agent is logged in.")
    heartbeat_interval: int = Field(15, description="The interval in seconds to send heartbeat requests.")
    poll_interval: int = Field(2, description="The interval in seconds to poll for new tasks.")
    request_timeout: int = Field(5, description="The timeout in seconds for HTTP requests.")

    def is_valid(self) -> bool:
        return all((self.token.get_secret_value(), self.server_host, self.tenant_domain))

    @property
    def server_url(self) -> str:
        if self.server_host.startswith("http"):
            return self.server_host
        return f"https://{self.server_host}"

    def set_auth_token(self, encoded_token: str):
        decoded = base64.urlsafe_b64decode(encoded_token.encode()).split(b"::")
        self.tenant_domain = decoded[0].decode()
        self.server_host = decoded[1].decode()
        self.token = SecretStr(base64.urlsafe_b64encode(decoded[2]).decode())

    def clear_auth_token(self):
        self.token = SecretStr("")

    @classmethod
    def load(cls, filename: Path | str | None = None) -> Self:
        if filename is None:
            filename = CONFIG_FILE_PATH
        filename = Path(filename)
        if not filename.exists():
            cfg = cls(server_host="", tenant_domain="", token="")
            cfg.save(filename)
        with open(filename) as f:
            content = f.read()
            return cls.model_validate_json(content)

    def save(self, filename: Path | None = None):
        if filename is None:
            filename = CONFIG_FILE_PATH

        filename.parent.mkdir(parents=True, exist_ok=True)
        with open(filename, "w") as f:
            content: dict = self.model_dump(mode="json")
            content["token"] = self.token.get_secret_value()
            f.write(json.dumps(content, indent=2))
            f.write("\n")


def parse_value(key: str, value: str) -> Any:
    field_info: "FieldInfo" = AgentConfig.model_fields[key]
    field_type = get_origin(field_info.annotation) or field_info.annotation
    if field_type is SecretStr:
        return SecretStr(value)
    if field_type is bool:
        return value.lower() in {"true", "yes", "y", "1"}
    return field_type(value)


CONFIG: AgentConfig = AgentConfig.load()



import importlib.metadata

try:
    VERSION = importlib.metadata.version("recurve-agent")
except importlib.metadata.PackageNotFoundError:
    # this should not happen in a normal environment, but it's useful for testing
    VERSION = "0.0.0"



from __future__ import annotations

import asyncio
from typing import Any, TypeVar, overload
from uuid import UUID

import httpx
from loguru import logger
from pydantic import BaseModel

from .._version import VERSION
from ..config import AgentConfig
from ..exceptions import APIError, MaxRetriesExceededException
from .schemas import Heartbeat, LoginPayload, Pagination, TaskItem, UpdateTaskInstanceStatus

ResponseModelType = TypeVar("ResponseModelType", bound=BaseModel)


class AgentClient:
    _config: AgentConfig
    _client: httpx.AsyncClient

    def __init__(self, config: AgentConfig):
        self.set_config(config)

    def set_config(self, config: AgentConfig):
        self._config = config
        self._client = httpx.AsyncClient(
            base_url=config.server_url,
            timeout=config.request_timeout,
            headers={"User-Agent": f"RecurveAgent/{VERSION}"},
        )

    @overload
    async def request(
        self, method: str, path: str, response_model_class: None = None, retries: int = 3, **kwargs
    ) -> Any:
        ...

    @overload
    async def request(
        self, method: str, path: str, response_model_class: type[ResponseModelType], retries: int = 3, **kwargs
    ) -> ResponseModelType:
        ...

    async def request(
        self,
        method: str,
        path: str,
        response_model_class: type[ResponseModelType] | None = None,
        retries: int = 1,
        **kwargs,
    ) -> Any:
        headers = kwargs.pop("headers", {})
        headers["Authorization"] = f"Bearer {self._config.token.get_secret_value()}"
        headers["X-Tenant-Domain"] = self._config.tenant_domain
        kwargs["headers"] = headers

        for attempt in range(retries):
            try:
                resp = await self._client.request(method, path, **kwargs)
                duration = resp.elapsed.total_seconds() * 1000
                logger.debug(f"[{method}] {resp.url} ({resp.status_code}), elapsed: {duration:.2f}ms, {kwargs}")

                resp.raise_for_status()
                resp_content = resp.json()

                # TODO(yangliang): handle errors more gracefully
                if resp_content["code"] != "0":
                    raise APIError(f"API request failed: {resp_content['msg']}")

                if response_model_class is not None:
                    return response_model_class.model_validate(resp_content["data"])
                return resp_content["data"]
            except httpx.HTTPStatusError as e:
                logger.debug(
                    f"HTTP error on attempt {attempt + 1} for url '{e.request.url}' :"
                    f" {e.response.status_code} - {e.response.text}"
                )
            except httpx.RequestError as e:
                logger.debug(f"Request error on attempt {attempt + 1} for url '{e.request.url}': {e}")

            if attempt < retries - 1:
                await asyncio.sleep(2**attempt)  # Exponential backoff
            else:
                raise MaxRetriesExceededException(
                    f"Failed to complete {method} request to {path} after {retries} attempts"
                )

    async def close(self):
        await self._client.aclose()

    async def healthcheck(self):
        return await self.request("GET", "/api/health")

    async def login(self, payload: "LoginPayload"):
        await self.request(
            "POST",
            "/api/agents/login",
            json=payload.model_dump(mode="json"),
            retries=1,
        )

    async def logout(self, agent_id: UUID):
        await self.request("POST", "/api/agents/logout", params={"agent_id": agent_id})

    async def heartbeat(self, payload: "Heartbeat"):
        await self.request(
            "POST",
            "/api/agents/heartbeat",
            json=payload.model_dump(mode="json"),
            retries=3,
        )

    async def get_pending_tasks(self, agent_id: UUID) -> Pagination[TaskItem]:
        result = await self.request(
            method="GET",
            path="/api/tasks/pending",
            params={"agent_id": agent_id},
            response_model_class=Pagination[TaskItem],
        )
        return result

    async def update_task_status(self, task_instance_id: int, payload: UpdateTaskInstanceStatus):
        await self.request(
            method="POST",
            path=f"/api/tasks/{task_instance_id}/status",
            json=payload.model_dump(mode="json"),
            retries=3,
        )
